(******************************************************************************)
(** * SRAG simulates SRAL: silent steps   *)
(******************************************************************************)

Require Import Lia.
Require Import Hahn.

Require Import Basic.
Require Import Subseq.

Require Import Labels Events Execution.
Require Import SRALn SRAG Subseq.
Require Import Write_lists Match_SRAn.

Set Implicit Arguments.

Lemma GsimL_dup (st st' : State)
          (G : execution) (MATCH: SGmatch st G)
          (DUP : forall (p : Tid) (ol' : list PStore),
      In ol' (st' p) ->
      exists (ol1 ol2 : list PStore) (_ : length ol2 > 0),
        In (ol1 ++ ol2) (st p) /\
        ol' = ol1 ++ (hd init_ps ol2 :: nil) ++ ol2)
(WF' : wf_state st') :
          SGmatch st' G.
Proof.
red; ins; splits; eauto.
all: try by eapply WF'.
all: try by eapply MATCH.
- red in WF'; desc.
ins. erewrite SAME_END; eauto.
- ins.

specialize (DUP _ _ IN); desf.
red in MATCH; desc.

specialize (SMATCH _ _ DUP); desc.
red in SMATCH; desc.
red in WL_WF; desc.
red in WL_CONS; desc.
exists (fun n => ifP n <= length ol1 then wl n
                 else wl (pred n) ).
red; splits; red; splits.
+ unfold event_in_wl, in_wl; unfolder; ins.
 apply WL_W.
autorewrite with calc_length in *; ins.
unfold event_in_wl, in_wl; unfolder; ins; desf.
 desf.
* 
  
  exists (wl n), x1.
  splits; eauto.
  exists n; splits; eauto.
  lia.
*  exists (wl (pred n)), x1.
  splits; eauto.
  exists (pred n); splits; eauto.
  lia.
+ unfold event_in_wl, in_wl; unfolder; ins.
 apply WL_G.
  autorewrite with calc_length in *; ins.
  unfold event_in_wl, in_wl; unfolder; ins; desf.
*  exists (wl n), x1.
  splits; eauto.
  exists n; splits; eauto.
  lia.
* exists (wl (pred n)), x1.
  splits; eauto.
  exists (pred n); splits; eauto.
  lia.
+ ins.
  autorewrite with calc_length in *. 
  rewrite nth_app; desf; try lia.
  desf.
* rewrite WL_TID; try lia.
  rewrite nth_app; desf; try lia.
  replace (n - length ol1) with 0 by lia; ins.
  unfold nth, hd; desf.
* rewrite WL_TID; try lia.
  rewrite nth_app; desf; try lia.
* rewrite WL_TID; try lia.
  rewrite nth_app; desf; try lia.
  replace (n - length ol1) with (S ((Nat.pred n - length ol1))) by lia; ins.
+ ins.
  autorewrite with calc_length in *. 
desf; rewrite WL_LOC; try lia; auto.

+ ins.
  autorewrite with calc_length in *. 
  rewrite nth_app; desf; try lia.
  desf.
* rewrite WL_VAL; try lia.
  rewrite nth_app; desf; try lia.
  replace (n - length ol1) with 0 by lia; ins.
  unfold nth, hd; desf.
* rewrite WL_VAL; try lia.
  rewrite nth_app; desf; try lia.
* rewrite WL_VAL; try lia.
  rewrite nth_app; desf; try lia.
  replace (n - length ol1) with (S ((Nat.pred n - length ol1))) by lia; ins.
+ ins.
  autorewrite with calc_length in *.
rewrite nth_app; desf; try lia.
* rewrite <- WL_RMW; try lia.
   rewrite nth_app; desf; try lia.
   replace (n - length ol1) with 0 by lia; ins.
   unfold nth, hd; desf.

* rewrite <- WL_RMW; try lia.
   rewrite nth_app; desf; try lia.
  replace (n - length ol1) with (S ((Nat.pred n - length ol1))) by lia; ins.
* rewrite <- WL_RMW; try lia.
  rewrite nth_app; desf; try lia.
+ remember (hb G)^? as rhb.
unfold event_in_wl, in_wl; unfolder; ins; desf.
* eapply WL_D.
  autorewrite with calc_length in *; ins.
remember (hb G)^? as rhb.
  unfold event_in_wl, in_wl; unfolder; ins; desf.
  splits; eauto 10.
  exists (wl n), x1.
  splits; eauto.
  exists n; splits; eauto.
  lia.
* eapply WL_D.
  autorewrite with calc_length in *; ins.
remember (hb G)^? as rhb.
  unfold event_in_wl, in_wl; unfolder; ins; desf.
  splits; eauto 10.
  exists (wl (pred n)), x1.
  splits; eauto.
  exists (pred n); splits; eauto.
  lia.
+ remember (hb G)^? as rhb.
  autorewrite with calc_length in *; ins.
unfold event_order_in_wl, worder_wl; unfolder; ins; desf; try lia.
* eapply WL_O.
  autorewrite with calc_length in *; ins.
remember (hb G)^? as rhb.
  unfold event_order_in_wl, worder_wl; unfolder; ins; desf.
  eexists; splits; eauto 10.
  exists (wl n1), (wl n2), x1.
  eexists; splits; eauto.
  exists n1, n2; splits; eauto.
  lia.
*  eapply WL_O.
  autorewrite with calc_length in *; ins.
remember (hb G)^? as rhb.
  unfold event_order_in_wl, worder_wl; unfolder; ins; desf.
  eexists; splits; eauto 10.
  exists (wl n1), (wl (pred n2)), x1.
  eexists; splits; eauto.
  exists n1, (pred n2); splits; eauto; lia.
* eapply WL_O.
  autorewrite with calc_length in *; ins.
remember (hb G)^? as rhb.
  unfold event_order_in_wl, worder_wl; unfolder; ins; desf.
  eexists; splits; eauto 10.
  exists (wl (pred n1)), (wl (pred n2)), x1.
  eexists; splits; eauto.
  exists (pred n1), (pred n2); splits; eauto; lia.
Qed.

Lemma GsimL_sub (st st' : State)
          (G : execution) (MATCH: SGmatch st G)
(SUB : forall (p : Tid) (ol' : list PStore),
      In ol' (st' p) ->
      exists ol : list PStore, In ol (st p) /\ subseq ol' ol)
(WF' : wf_state st') :
          SGmatch st' G.
Proof.
red; ins; splits; eauto.
all: try by eapply WF'.
all: try by eapply MATCH.
- red in WF'; desc.
ins. erewrite SAME_END; eauto.
- ins.

specialize (SUB _ _ IN); desf.
rename ol into ol'.
rename ol0 into ol.
apply subseq_mapping in SUB0; desc.
red in MATCH; desc.
specialize (SMATCH _ _ SUB); desc.
exists (fun n => wl (f n)).
red in SMATCH; desc.
red in WL_WF; desc.
red in WL_CONS; desc.
red in SUB0; desf.
red; splits; red; splits.
+ unfolder; ins.
  eapply WL_W.
  unfold event_in_wl, in_wl in *; desf.
  eexists _,_; splits; eauto.
+ unfolder; ins.
  eapply WL_G.
  unfold event_in_wl, in_wl in *; desc.
  eexists _,_; splits; eauto.
+ ins.
  rewrite WL_TID; eauto.
  erewrite nth_error_nth; try edone.
  rewrite <- NTH.
  by apply nth_error_nth_len.
+ ins.
  rewrite WL_LOC; eauto.
+ ins.
  rewrite WL_VAL; eauto.
  erewrite nth_error_nth; try edone.
  rewrite <- NTH.
  by apply nth_error_nth_len.
+ ins.
  rewrite <- WL_RMW; eauto.
  cut ((nth n ol' init_ps x) = (nth (f n) ol init_ps x)).
by intro A; rewrite A; vauto.
  erewrite nth_error_nth; try edone.
  rewrite NTH.
  apply nth_error_nth_len.
eauto.
+ unfolder; ins; desc.
  unfold event_in_wl, in_wl in *; desc.
  eapply WL_D.
  unfolder; splits; eauto 12.
+ unfolder; ins; desc.
  unfold event_order_in_wl, worder_wl in *; desc.
  eapply WL_O.
  unfolder; splits; subst.
eexists; splits; eauto.
eexists (wl (f n1)),(wl (f n2)),_,_; splits; eauto.
eexists (f n1),(f n2); splits; eauto.
apply Lt.le_lt_or_eq_stt in H; destruct H; subst.
apply NPeano.Nat.lt_le_incl.
by apply MON.
lia.
Qed.

Lemma GsimL_silent (st st' : State)
          (G : execution) (MATCH: SGmatch st G)
          (SRAL: SRAL_step st st' None) :
          SGmatch st' G.
Proof.
inversion SRAL; subst; clear SRAL.
eby eapply GsimL_dup.
eby eapply GsimL_sub.
Qed.
