(******************************************************************************)
(** * Definition of labels *)
(******************************************************************************)

Require Import FinFun.
Require Import Hahn.

Set Implicit Arguments.

(******************************************************************************)
(** ** Assume a finite non-empty set of location  *)
(******************************************************************************)

Parameter Loc : Type.
Parameter Loc_list : list Loc.
Axiom Loc_list_prop : Full Loc_list.
Axiom X0 : Loc.

(******************************************************************************)
(** ** Assume a finite non-empty set of thread ids  *)
(******************************************************************************)

Parameter Tid : Type.
Parameter Tid_list : list Tid.
Axiom Tid_list_prop : Full Tid_list.
Axiom T0 : Tid.

(******************************************************************************)
(** ** Labels  *)
(******************************************************************************)

Section Labels.

Definition Val := nat.

Inductive Lab := 
  | Aload  (x:Loc) (v:Val)
  | Astore (x:Loc) (v:Val)
  | Armw   (x:Loc) (vr vw:Val).

Definition loc_l l :=
  match l with
  | Aload  x _
  | Astore x _
  | Armw   x _ _ => x
  end.

Definition valr_l l :=
  match l with
  | Aload  _ v
  | Armw   _ v _ => v
  | _ => 0 
  end.

Definition valw_l l :=
  match l with
  | Astore _ v
  | Armw   _ _ v => v
  | _ => 0 
  end.

Definition is_r_l l := 
  match l with
  | Aload  _ _ => true
  | Armw   _ _ _ => true
  | _ => false
  end.

Definition is_w_l l := 
  match l with
  | Astore  _ _  => true
   | Armw   _ _ _ => true
  | _ => false
  end.

Definition is_rmw_l l := 
  match l with
  | Armw   _ _ _ => true
  | _ => false
  end.

Definition default_lab := Astore X0 0.

End Labels.

