(******************************************************************************)
(** * SRAL simulates SRAG: read step   *)
(******************************************************************************)

Require Import List.
Import ListNotations.

Require Import PeanoNat Lt Gt.
Require Import Lia.
Require Import Hahn.

Require Import AuxRel Subseq.

Require Import Basic.

Require Import Labels Events Execution.
Require Import SRALn Wmax Wmax_t SRAG.
Require Import Write_lists Match_SRAn.

Require Import FG.

Set Implicit Arguments.

Lemma LsimG_read_wl_helper (G G' : execution)
  t n w l (ADD : add G G' t n l w) (NW : ~ is_w_l l)
  p ol wl (MATCH' : SGmatch_wl G' p ol wl)
  (WL: ~ event_in_wl wl (length ol) (ThreadEvent t n l))  :
SGmatch_wl G p ol wl.
Proof.
red in MATCH'; desc.
red in WL_WF; desc.
red in WL_CONS; desc.
red; unfold wf_wl,consistent_wl; unnw.
rewrite (add_co_monotone ADD), (add_hb_monotone ADD).
splits; try edone.
- rewrite (add_acts_set ADD) in WL_G.
  unfolder; intros a INWL.
  specialize (WL_G a INWL).
  unfolder in WL_G; desf.
- ins; split; ins.
  + rewrite (add_co_monotone ADD). 
    by apply WL_RMW.
  + eapply WL_RMW; try edone.
    rewrite add_non_write_co; try edone.
Qed.

Lemma LsimG_read_wl (G G' : execution)
  t n x v w (ADD : add G G' t n (Aload x v) w)
  p ol wl (MATCH' : SGmatch_wl G' p ol wl) :
  SGmatch_wl G p ol wl.
Proof.
 eapply LsimG_read_wl_helper; try edone.
    intro INwl.
    red in MATCH'; desc.
    red in WL_WF; desc.
    specialize (WL_W _ INwl).
    desf.
Qed.

Definition make_ps (f : Loc -> Event) G := 
  fun y => RT (tid (f y)) (valw (f y)) 
  (ifP (w_max G y (f y)) then true else false).

Lemma LsimG_read_wl_t (G G' : execution) (WF_G: Wf G)
  t n x v w (ADD : add G G' t n (Aload x v) w)
  L (LEN : length L > 0) wl (MATCH' : SGmatch_wl G' t L wl)
  (CO : ~ dom_rel (co G ⨾ (hb G)^? ⨾ ⦗is_tid t⦘) w)
  f (WMAX: forall y, w_max_t G' t y (f y)) :
  SGmatch_wl G t (make_ps f G :: L) (extend_wl wl f).
Proof.
red in MATCH'; desc.
red in WL_WF; desc.
red in WL_CONS; desc.
red; splits; red; ins; splits; desc; simpl. 
- rewrite event_in_wl_extend_wl; try done.
  rewrite WL_W. 
  unfolder; ins; desf.
  apply WMAX; done.
- rewrite event_in_wl_extend_wl; try done.
  arewrite (event_in_wl wl (length L) ⊆₁ 
  event_in_wl wl (length L) ∩₁ event_in_wl wl (length L)).
  rewrite WL_W at 1.
  rewrite WL_G at 1.
  arewrite ((fun e => exists x0, f x0 = e) ⊆₁ 
    (fun a : Event => W a) ∩₁ acts_set G').
  by unfolder; ins; splits; desf; apply WMAX; done.
  rewrite (add_acts_set ADD).
  revert WL_G; unfolder; ins; desf.
- destruct L; [ins; lia|]; ins.
  unfold extend_wl, make_ps; desf; ins; apply WL_TID; lia.
- destruct L; [ins; lia|]; ins.
  unfold extend_wl, make_ps; desf; ins; subst.
  apply WMAX; done.
  apply WL_LOC; lia.
- destruct L; [ins; lia|]; ins.
  unfold extend_wl, make_ps; desf; ins; apply WL_VAL; lia.
- (* destruct L; [ins; lia|]; ins. *)
  ins. destruct n0.
  + simpl; unfold extend_wl, make_ps; ins; desf.
    split; ins; try apply w0.
    split; ins.
    exfalso.
    apply n0.
    unfold w_max; unfolder; ins; desf; eauto.
    splits; eauto; try by apply WMAX.
    enough (acts_set G' (f x0)).
    { apply (add_acts_set ADD) in H0; unfolder in H0; desf.
      exfalso.
      specialize (WMAX x0).
      rewrite <- H0 in WMAX.
      unfold w_max_t in WMAX. 
      unfold set_compl in WMAX; unfolder in WMAX; desf; ins. }
    apply WMAX.
  +  unfold extend_wl; ins.
     split; ins.
     * rewrite (add_co_monotone ADD).
       eapply WL_RMW with (n:=n0). lia. eauto.
     * apply WL_RMW; try lia.
       rewrite (add_non_write_co); try edone. type_solver.
- rewrite event_in_wl_extend_wl; auto.
  relsf; splits.
  + by rewrite (add_co_monotone ADD), (add_hb_monotone ADD).
  + unfolder; ins; desc. 
    eapply WMAX.
    generalize (add_co_monotone ADD), (add_hb_monotone ADD).
    basic_solver 21.
- rewrite event_order_in_wl_extend_wl; relsf.
  repeat (apply irreflexive_union; splits).
  + rewrite (add_co_monotone ADD), (add_hb_monotone ADD); done.
  + rewrite <- !seqA in WL_D.
    rewrite add_non_write_co_hb in WL_D; try edone.
    2: basic_solver.
    unfolder; ins. desc; subst.
    specialize (WMAX x1).
    red in WMAX; unfold set_compl in WMAX; unfolder in WMAX; desc.
    destruct WMAX1. 
    desc; subst. 
    eapply WL_D; unfolder; splits; eauto.
    desf; eauto 21.
    * eapply add_read_hb in H; vauto.
      unfolder in H; desf; eauto 21.
      apply wf_co_init in H0; unfolder in H0; desf.
      apply no_hb_to_init in H; unfolder in H; desf.
    * eapply add_read_hb in H; vauto.
      generalize (@hb_trans G).
      unfolder in H; desf; eauto 21.
      apply wf_co_init in H0; unfolder in H0; desf.
      apply no_hb_to_init in H1; unfolder in H1; desf.
      apply no_hb_to_init in H; unfolder in H; desf.
    * desf.
      apply wf_co_init in H0; unfolder in H0; desf.
      apply no_hb_to_init in H1; unfolder in H1; desf.
  + unfolder; ins; desc; subst.
    assert (WMAXx1 := WMAX x1).
    specialize (WMAX x2).
    red in WMAX; unfold set_compl in WMAX; unfolder in WMAX; desc.
    destruct WMAX1. 
    * desc; subst. 
      eapply WMAXx1; unfolder; splits; eauto.
      apply (add_co_monotone ADD) in H0.
      assert (z0 = f x2 \/ hb G' z0 (f x2)).
      by generalize (add_hb_monotone ADD); basic_solver.
      clear H1.
      generalize (@hb_trans G'); basic_solver 21.
    * desf.
      apply wf_co_init in H0; unfolder in H0; desf.
      apply no_hb_to_init in H1; unfolder in H1; desf.
Qed.

Lemma LsimG_read (st' : State) (WF_S: wf_state st')
 (G G' : execution) (WF_G: Wf G) (CONS: sra_consistent G)
 t x v (SRAG: SRAG_step G G' (Some (t, (Aload x v))))
 (MATCH': SGmatch st' G') : 
 exists st, SRAL_step st st' None /\
            SRAL_step st st (Some (t, (Aload x v))) /\ 
            SGmatch st G.
Proof.
assert (WF_G' : Wf G').
eapply SRAG_step_preserves_wf; try edone.
inversion SRAG; subst; clear SRAG; desf; ins.
rename x0 into t.
forward (apply (max_w_t_exists_unique_fun' WF_G' t)).
ins; desc.
exists (fun p => ifP (p = t) then 
  map (cons (make_ps f G)) (st' t)
 else st' p).
clear MAX.
splits.
- apply step_lose; [|by eapply SGmatch_implies_wf_st, MATCH'].
  ins. desf.
  2: by eexists; splits; eauto; apply subseq_reflexive.
  eexists. splits.
  apply in_map; apply IN.
  econs. apply subseq_reflexive.
- eapply step_read; try done.
  red. splits; desf.
  desf.
  exists (tid (f x)). ins.
  eapply in_map_iff in IN; desf; ins; desf.
  unfold make_ps; f_equal.
  specialize (H x).
  cdes ADD; desc; ins.
  rewrite <- VALw; auto.
  f_equal.
  eapply max_w_t_unique with (G:=G'); try edone.
  eapply w_max_t_add_read_same_loc.
  6: eapply ADD. all: eauto.
- red; splits; eauto.
  + red in WF_S; desc. 
    ins; desf.
    specialize (NON_EMPTY_POT t).
    destruct (st' t); desf.
  + ins; desf.
    eapply in_map_iff in IN; desc; desf.
    eby eapply WF_S.
  + ins; desf; [|eby eapply MATCH'].
    eapply in_map_iff in IN; desc; desf. 
    rename x0 into D. ins.
    destruct (@length PStore D) eqn:E; ins.
    by destruct D; desf; ins; exfalso; eapply WF_S; eauto.
    replace n0 with (pred (@length PStore D)) in * by lia.
    eby eapply MATCH'.
  + ins; desf.
    * red in MATCH'; desc.
      rewrite in_map_iff in IN; desf.
      specialize (SMATCH _ _ IN0); desf.
      eapply LsimG_read_wl_t in SMATCH; eauto.
      apply NON_EMPTY_OL in IN0.
      destruct x0; ins.
      lia.
    * red in MATCH'; desc.
      specialize (SMATCH _ _ IN); desc.
      eexists.
      eapply LsimG_read_wl; try edone.
Qed.
