(******************************************************************************)
(** * SRAL simulates SRAG: write step   *)
(******************************************************************************)

Require Import List.
Import ListNotations.


Require Import PeanoNat Lt Gt.
Require Import Lia.
Require Import Hahn.
Require Import Init.Nat.


Require Import Coq.Logic.ClassicalChoice.

Theorem choice2 :
 forall (A1 A2 B : Type) (R : A1->A2->B->Prop),
   (forall (x1 : A1) (x2 : A2), exists y : B, R x1 x2 y) ->
    exists f : A1->A2->B, (forall (x1 : A1) (x2 : A2), R x1 x2 (f x1 x2)).
Proof.
ins.
assert (forall (x1 : A1), exists f : A2 -> B, 
        forall (x2 : A2), R x1 x2 (f x2)).
{ ins; eapply choice in H; edone. }
by apply choice in H0.
Qed.

Theorem choice3 :
 forall (A1 A2 B : Type) (P : A1 -> A2 -> Prop) 
   (R : A1->A2->B->Prop) (d : B),
   (forall (x1 : A1) (x2 : A2) (IN_P: P x1 x2), 
    exists y : B, R x1 x2 y) ->
    exists f : A1->A2->B, 
    forall (x1 : A1) (x2 : A2) (IN_P: P x1 x2),
    R x1 x2 (f x1 x2).
Proof.
ins.
set (Q:= fun x1 x2 x3 => P x1 x2 -> R x1 x2 x3).
cut (exists f : A1 -> A2 -> B,
  forall (x1 : A1) (x2 : A2),Q x1 x2 (f x1 x2)).
{ ins; desc. }


eapply choice2 with (A1:=A1) (A2:=A2) (B:=B).
unfold Q in *; ins.
specialize (H x1 x2).
destruct (classic (P x1 x2)).
specialize (H H0); desc; eauto.
exists d; ins.
Qed.

Require Import Basic.

Require Import Labels Events Execution.
Require Import SRALn Wmax Add_max SRAG.
Require Import Write_lists Match_SRAn.

Require Import FG.

Set Implicit Arguments.



Definition max_w_before_set G Wset x := 
  acts_set G ∩₁ W ∩₁ (fun w => loc w = x) ∩₁ 
  (dom_rel ((hb G)^? ⨾ ⦗Wset⦘)) ∩₁ 
  (max_elt (G.(co) ;; ⦗dom_rel ((hb G)^? ⨾ ⦗Wset⦘)⦘)).

Lemma max_w_before_set_unique G (WF: Wf G) Wset x w1 w2 
  (MAX1: max_w_before_set G Wset x w1) 
  (MAX2: max_w_before_set G Wset x w2) : 
  w1 = w2.
Proof.
destruct (classic (w1 = w2)); try done.
exfalso.
unfold max_w_before_set in *; desc.
unfold max_elt in MAX1.
unfold max_elt in MAX2.
unfolder in * ; desc.
eapply (wf_co_total WF) in H; desf; eauto 10.

all: try by apply (wf_co_init WF) in H; unfolder in H; desf.


Qed.

Lemma max_w_before_set_exists G (WF: Wf G) (Wset : Event -> Prop) x 
  (INIT : Wset (InitEvent x)) : 
  exists w, max_w_before_set G Wset x w.
Proof.
assert (T: transitive (co G ⨾ ⦗dom_rel ((hb G)^? ⨾ ⦗Wset⦘)⦘)).
 { generalize (co_trans WF). basic_solver 12. } 

assert (exists w, (G.(co) ;; ⦗dom_rel ((hb G)^? ⨾ ⦗Wset⦘)⦘)＊ (InitEvent x) w  
/\ max_elt (G.(co) ;; ⦗dom_rel ((hb G)^? ⨾ ⦗Wset⦘)⦘)  w).
{ eapply last_exists with (dom := G.(acts)).
  eapply acyclic_mon.
  eby eapply co_acyclic.
  basic_solver.
  intros c H.
  apply rt_of_trans in H; try edone.
  hahn_rewrite (dom_r (wf_coE WF)) in H.
  unfolder in H; desf.
  by apply WF. }
desc.
apply clos_refl_trans_of_transitive in H; try edone.
assert (acts_set G w).
{ destruct H.
  by subst; eapply WF.
  hahn_rewrite (dom_r (wf_coE WF)) in H; unfolder in H.
  unfold acts_set in *; desf. }
assert (W w).
{ destruct H.
  subst; type_solver.
  hahn_rewrite (dom_r (wf_coD WF)) in H.
  unfolder in H; desf. }
assert (loc w = x).
{ destruct H.
  subst; type_solver.
  hahn_rewrite WF.(wf_col) in H.
  unfold same_loc in *; unfolder in H. desf. }
exists w. red. unfold max_elt; unfolder. splits; try done.
unfolder in H; desf; eauto 20.
Qed.

Lemma max_w_before_sets_exists 
  G (WF : Wf G) N (wSets : nat -> (Event -> Prop)) x 
    (INIT : forall n (INWS: n < N), wSets n (InitEvent x)) :
  exists wSeq, << WSEQ: forall n (INWS: n < N),
    max_w_before_set G (⋃₁ k <= n , wSets k) x (wSeq n) >>.
Proof.
induction N.
- exists (fun _ => (InitEvent x)).
  unnw; ins; lia.
- forward apply max_w_before_set_exists with (Wset := ⋃₁k <= N, wSets k); eauto.
  by unfolder; eauto.
  forward apply IHN.
  ins; apply INIT; lia.
  ins; desc.
  exists (upd wSeq N w); splits; ins; eauto.
  unfold upd; desf.
  apply WSEQ.
  lia.
Qed.

Lemma max_w_before_sets_exists_last 
  G (WF : Wf G) N (wSets : nat -> (Event -> Prop)) x 
    (INIT : forall n (INWS: n < N), wSets n (InitEvent x)) :
  exists wSeq, << WSEQ: forall n (INWS: n < N),
    max_w_before_set G (⋃₁ k <= n , wSets k) x (wSeq n) >> /\ 
  << LAST_WSEQ: w_max G x (wSeq N)>>.
Proof.
forward apply max_w_before_sets_exists; eauto; ins; desf.
forward apply max_w_exists; eauto; ins; desf.
exists (upd wSeq N w).
unnw; unfold upd; splits; ins; desf; eauto; lia.
Qed.

Lemma first_index_exists (A : Type) (f : nat -> A) N (Z: N > 0) (tst : A -> Prop) 
(NE: exists m, m < N /\ tst (f m))
(STABLE: forall n m, n <= m < N -> tst (f n) -> tst (f m)) :
exists i, i < N /\ tst (f i) /\ forall j, j < i -> ~ tst (f j).
Proof.
generalize dependent f.
induction N.
lia.
ins; desc.
destruct (classic (tst (f 0))).
by exists 0; splits; auto; lia.
destruct m; desf.
destruct N; ins; try lia.
forward apply IHN with (f:=fun n => f (S n)).
lia.
by exists m; splits; try lia; auto.
ins; eapply STABLE with (n:= S n); auto; lia.
ins; desc.
exists (S i); splits; try lia.
ins.
destruct j; auto.
specialize (H2 j); ins.
apply H2.
lia.
Qed.

Lemma LsimG_write_main (st': State) (G G': execution) 
          t l (SRAG: SRAG_step G G' (Some (t, l)))
          x v (LAB_W : is_w_l l) 
          (LOC : loc_l l = x) (VAL : valw_l l = v)
          (MATCH': SGmatch st' G') : 
          exists st, SRAL_write_step_helper st st' t x v /\ 
                     SGmatch st G /\
                     ((is_r_l l = true) -> 
forall L (IN: In L (st t)), rmw_flag_opt ((nth 0 L init_ps) x) = true).
Proof.
inversion SRAG; subst; clear SRAG; desf; ins.
specialize (MAX hahn__true_is_true).
clear CO.
assert (MATCH'' := MATCH').
rename x0 into t.
rename n into x0.
rename l0 into l.
remember (loc_l l) as x.
remember (valw_l l) as v.
unfold SGmatch; ins.
red in MATCH''; desc.
generalize (SGmatch_implies_wf_st MATCH'); intro WF_ST.
cut (forall p ol' (IN: In ol' (st' p)), 
exists (ol : list PStore) i,
  << I_LEN: i < length ol >> /\ 
  << I_T: p = t -> i = 0 >> /\
  << RMW: rmw_flag_opt (nth (pred (length ol)) ol init_ps x) = true >> /\
  << UPD: ol' = update_ol ol i x t v >> /\
  << MATCH_T: exists wl_t, 
    SGmatch_wl G t (skipn i ol) wl_t /\
    ((is_r_l l = true) -> max_elt G.(co) (wl_t 0 x)) >> /\
  << MATCH_P: exists wl_p, SGmatch_wl G p ol wl_p >>).
(*   << RMW_T: (is_r_l l = true) -> rmw_flag_opt (nth 0 (skipn i ol) init_ps x) = true >>). *)
{ ins; desc.
  apply choice3 in H; desc; try apply []. 
  rename f into get_ol.
  apply choice3 in H; desc; try apply 0.
  rename f into get_i.

assert (SAME_LEN : forall p ol' (IN: In ol' (st' p)), 
 length  (get_ol p ol') = length ol').
{ ins.
  specialize (H _ _ IN); desf.
  rewrite UPD at 2.
  unfold update_ol.
  autorewrite with calc_length in *; ins.
  rewrite skipn_length; lia. }

assert (NY: forall p ol' (IN: In ol' (st' p)) n y (NX: y <> x), 
 nth n ol' init_ps y = nth n (get_ol p ol') init_ps y).
{ ins.
  specialize (H _ _ IN); desf.
  rewrite UPD at 1.
  unfold update_ol.
  rewrite nth_app; desf; 
  remember (loc_l l) as x;
  remember (valw_l l) as v.
  - autorewrite with calc_length in *; ins.
    rewrite <- skipn_map, skipn_nth.
    replace (get_i p ol' + (n - min (get_i p ol') (length (get_ol p ol')))) with n by lia.
    destruct (classic (n < length (get_ol p ol'))).
    + rewrite nth_indep with (d':= upd_ps init_ps x t v).
      rewrite map_nth with (f:= fun ps : PStore => upd_ps ps x t v).
      unfold upd_ps, upd; desf.
      autorewrite with calc_length in *; ins.
    + rewrite !nth_overflow; try done; 
      autorewrite with calc_length in *; ins; try lia.
  - erewrite nth_indep with (d':= upd_read_ps init_ps x); try apply l.
    rewrite map_nth with (f:= fun ps : PStore => upd_read_ps ps x).
    unfold upd_read_ps, upd; desf.
    rewrite nth_firstn; try done.
    autorewrite with calc_length in *; ins.
    lia. lia. }

set (st := fun p : Tid => (ifP (p = t) then 
flatten (map (fun p' => map (fun ol => skipn (get_i p' ol) (get_ol p' ol)) (st' p')) Tid_list)
else []) ++ map (get_ol p) (st' p)).

exists st; splits; ins; auto; unfold st in *; clear st.
- red; splits; ins; desf.
  + red; ins.
    specialize (H t ol' IN); desc.
    exists (get_ol t ol'), (get_i t ol').
    splits; eauto.
    * apply in_app_r. apply in_map. done.
    * apply in_app_l. in_simp.
      eexists; splits.
      in_simp; eauto. 
      exists t; splits; eauto.
      by apply Tid_list_prop.
      in_simp; eauto.
  + red; ins.
    specialize (H p ol' IN); desc.
    exists (get_ol p ol'), (get_i p ol').
    splits; eauto.
    by apply in_map.
    apply in_app_l. in_simp.
    eexists; splits.
    * in_simp. exists p; splits; eauto.
      by apply Tid_list_prop.
    * in_simp; eauto.
- desf.
  + intros X.
    apply app_eq_nil in X; desc.
    apply map_eq_nil in X0; desc.
    specialize (NON_EMPTY_POT t).
    tauto.
  + intros X.
    specialize (NON_EMPTY_POT t0).
    apply NON_EMPTY_POT; auto.
    eapply map_eq_nil; eauto.
- in_des; desf. in_des.
  + apply skipn_nil2 in IN0; desf.
    * specialize (H _ _ IN2); desc. 
      rename x into m. 
(* rename x1 into x1. *)
      assert (get_i m x1 < length (get_ol m x1)) by eapply I_LEN.
      lia.
    * eapply NON_EMPTY_OL; try edone.
      specialize (H _ _ IN2); desf.
      rewrite UPD, IN0.
      unfold update_ol.
      rewrite firstn_nil, skipn_nil; ins.
  + specialize (H t0 x1 IN0); desf.
    rewrite H0 in I_LEN; ins; lia.
- destruct (classic (y = x)); subst. 
  + desf; in_des; rewrite ?skipn_length;
    try specialize (H _ _ IN1); 
    try specialize (H _ _ IN2);
    desf; try lia.
    all: try rewrite skipn_nth.
    all: rename x into m.
(*  try rename x2 into x1. *)
    all: try replace (get_i m x1 + pred (length (get_ol m x1) - get_i m x1)) 
     with (pred (length (get_ol m x1))) by lia.
    by auto.
    by eapply H.
    by eapply H.
  + desf; in_des; rewrite ?skipn_length;
    try specialize (H _ _ IN1); 
    try specialize (H _ _ IN2);
    desf; try lia.
    all: try rewrite skipn_nth.
    all: rename x into m. 
(* try rename x2 into x1. *)
    all: try replace (get_i m x1 + pred (length (get_ol m x1) - get_i m x1))
     with (pred (length (get_ol m x1))) by lia.
    all: rewrite <- NY; eauto.
    all: rewrite SAME_LEN.
    all: try erewrite RMW_END; eauto.
- desf; ins; in_simp.
  + rewrite in_app_iff,in_flatten_iff in IN; desf; in_simp; ins; eauto.
    by specialize (H _ _ IN2); desc; eauto.
    by specialize (H _ _ IN0); desc; eauto.
  + specialize (H _ _ IN0); desc; eauto.
- desf; ins.
  rewrite in_app_iff,in_flatten_iff in IN; desf; in_simp; ins; eauto.
  + specialize (H _ _ IN2); desc.
red in MATCH_T; desc. red in WL_WF; desc.


apply WL_RMW; auto.
  autorewrite with calc_length in *; ins.
destruct (skipn (get_i x x1) (get_ol x x1)) eqn:E; ins; try lia.
exfalso; apply skipn_nil2 in E; desf.
lia.
rewrite E in I_LEN; ins; lia.

+ 
  specialize (H _ _ IN0); desc.
red in MATCH_T; desc. red in WL_WF; desc.

rewrite I_T in WL_RMW; ins.
apply WL_RMW.
lia. auto.
 }

ins.
clear MATCH'.
specialize (SMATCH _ _ IN); desc.
rename wl into wl'.

assert (OL'LEN : length ol' > 0).
{ destruct ol'; ins; try lia.
  exfalso; eapply NON_EMPTY_OL; eauto. }

clear NON_EMPTY_POT.
clear NON_EMPTY_OL.
clear WF_ST.

assert (WMAX: forall e, 
           w_max G' x e <-> e = ThreadEvent t x0 l).
{ ins; split; ins.
  - eapply max_w_unique; eauto.
    eapply w_max_add_write' with (G:=G) (w:=w); try edone.
  - subst.
    eapply w_max_add_write with (G:=G) (w:=w); try edone. } 

assert (WMAX_HELPER': forall x j (LEQ: j < length ol'), 
  w_max G' x (wl' j x) <-> max_elt G'.(co) (wl' j x)).
{ unfold w_max; ins; unfolder; split; ins; desf; eauto.
  splits; eauto.
  apply SMATCH; apply event_in_wl_k; lia.
  apply SMATCH; apply event_in_wl_k; lia.
  red in SMATCH; desc; red in WL_WF; desc.
  rewrite WL_LOC; auto. }

assert (WMAX_HELPER: forall y (NEQ: y <> x) j (LEQ: j < length ol'), 
  w_max G y (wl' j y) <-> max_elt G.(co) (wl' j y)).
{ unfold w_max; ins; unfolder; split; ins; desf; eauto.
  splits; eauto.
  - cut (acts_set G' (wl' j y)).
    { ins; eapply add_acts_set in H0; try edone.
      unfolder in H0; desf. exfalso. apply NEQ.
      red in SMATCH; desc. 
      red in WL_WF; desc.
      erewrite <- WL_LOC with (x:=y).
      rewrite <- H0. ins.
      lia. }
    apply SMATCH.
    apply event_in_wl_k; lia.
  - apply SMATCH.
    apply event_in_wl_k; lia.
  - red in SMATCH; desc; red in WL_WF; desc.
    rewrite WL_LOC; auto. }

assert (T0 : p = t -> w_max G' x (wl' 0 x)).
{ ins; subst.
  red in SMATCH; desc.
  red in WL_CONS; desc.
  apply WMAX_HELPER'; try lia.
  unfolder; splits.
  ins.
  eapply WL_D with (x := (wl' 0 (loc_l l))); unfolder; splits.
  apply event_in_wl_k; lia.
  exists ((ThreadEvent t x0 l)),(ThreadEvent t x0 l); splits; try done.
  2: eexists; splits; vauto.
  forward eapply max_w_maximum with (G:=G') (w := (wl' 0 (loc_l l))); try edone.
  - apply WL_WF. apply event_in_wl_k; lia.
  - apply WL_WF. apply event_in_wl_k; lia.
  - eapply w_max_add_write with (G:=G); try edone; vauto.
    replace (loc (wl' 0 (loc_l l))) with (loc_l l); try done.
    by symmetry; apply WL_WF.
  - unfolder; ins; desf.
    rewrite H in REL.
    exfalso.
    forward eapply w_max_add_write with (G:=G); try edone; vauto.
    unfold w_max; unfolder; ins; desf; eauto. }

assert (I_HELPER : forall k j, k <= j < length ol' -> 
  w_max G' x (wl' k x) -> w_max G' x (wl' j x)).
{ ins; subst. apply WMAX_HELPER'; try lia.
  red in SMATCH; desc.
  red in WL_CONS; desc.
  unfolder; splits.
  eapply WL_O with (x := (wl' k (loc_l l))); unfolder; splits.
  unfold event_order_in_wl, worder_wl; splits.
  eexists (wl' j (loc_l l)); splits.
  eexists _,_; splits; eauto 10.
  exists (wl' k (loc_l l)); splits; eauto.
  forward eapply max_w_maximum with (G:=G') (w := (wl' j (loc_l l))); try edone.
  - apply WL_WF. apply event_in_wl_k; lia.
  - apply WL_WF. apply event_in_wl_k; lia.
  - replace (loc (wl' j (loc_l l))) with (loc_l l); try edone.
    symmetry; apply WL_WF; lia.
  - unfolder; ins; desf; eauto.
    exfalso. rewrite <- H2 in H0.
    red in H0; unfolder in H0; desf; eauto. }

assert (I: exists i, i < length ol' /\ w_max G' x (wl' i x) /\ 
  forall j, j < i -> ~ (w_max G' x) (wl' j x)).
{ eapply first_index_exists; auto.
  exists (pred (length ol')); splits; try lia.
  eapply wf_wl_last_max; try edone.
  apply SMATCH.
  lia. erewrite RMW_END; ins; eauto. }

desc; subst.

set (Wsets := fun n => 
  dom_rel (⦗is_tid p ∪₁ is_tid t ∪₁ is_init ∪₁
  (fun e => exists y, e = wl' n y)⦘) 
  ∪₁ ((fun e => is_r_l l = true) ∩₁ eq w)).

forward apply (@max_w_before_sets_exists_last G WF_G (pred (length ol')) Wsets (loc_l l)).
{ ins; unfold Wsets. left. unfolder; unfold is_init; eauto 21. }
ins; desc.

remember (loc_l l) as x.
set (ol := mk_list i (fun n => ifP (w_max G x (wl' n x)) then
   upd (nth n ol' init_ps) x 
  (RT (tid_opt (nth n ol' init_ps x)) (val_opt (nth n ol' init_ps x)) true)
  else (nth n ol' init_ps)) ++
  mk_list (length ol' - i) (fun n => upd (nth (n + i) ol' init_ps) x 
  (RT (tid (wSeq (n + i))) (valw (wSeq (n + i))) 
  (ifP (w_max G x (wSeq (n + i))) then true else false)))).

assert (LEN' : @length PStore ol = length ol').
{ unfold ol.
  rewrite !length_app, !length_mk_list.
  lia. }

exists ol, i; splits; try lia.


- ins; subst. destruct i; ins; desf.
  exfalso; eapply I1 with (j:=0); try lia; auto. 
- rewrite LEN'. unfold ol in *. clear ol. ins. 
  autorewrite with calc_length in *; ins.
  rewrite nth_app; desf;
  autorewrite with calc_length in *; ins; try lia.
  rewrite nth_mk_list; desf;
  autorewrite with calc_length in *; ins; try lia.
  unfold upd; desf.
  unfold upd; desf.
  ins.
  replace ((pred (length ol') - i + i)) with (pred (length ol')) in * by lia.
  eauto.
- unfold ol in *.
  unfold update_ol; ins.
  clear ol.
  rewrite firstn_app.
  replace (i - length (firstn i ol')) with 0; ins.
  2: { rewrite length_firstn. lia. }
  autorewrite with calc_length in *; ins; try lia.
  rewrite Nat.sub_diag; ins.
  rewrite app_nil_r, firstn_mk_list; try lia.
  rewrite skipn_app.
  autorewrite with calc_length in *; ins.
  rewrite Nat.sub_diag; ins.
  rewrite map_app, skipn_mk_list; try edone.
  2: apply (fun _ : Loc => init_ps).
  rewrite !map_mk_list, !Nat.sub_diag; ins.
  unfold upd_read_ps, upd_ps, upd.
  eapply nth_ext with (d:=init_ps).
  by autorewrite with calc_length in *; ins; lia.
  ins.
  rewrite !nth_app.
  autorewrite with calc_length in *; ins.
  rewrite !nth_mk_list; desc; try lia; ins.
  apply functional_extensionality; ins; desf; try lia.
  all: try replace (n - i + i) with n in * by lia; auto; ins.
  + red in SMATCH; desc.
    red in WL_WF; desc.
    destruct (nth n ol' init_ps (loc_l l)) eqn:E.
    f_equal.
    * specialize (WL_TID n H (loc_l l)). rewrite E in WL_TID; ins.
      rewrite <- WL_TID.
      forward eapply I_HELPER with (k:=i) (j:=n); try lia; auto.
      ins; rewrite WMAX in *; rewrite H0; vauto.
    * specialize (WL_VAL n H (loc_l l)). rewrite E in WL_VAL; ins.
      rewrite <- WL_VAL.
      forward eapply I_HELPER with (k:=i) (j:=n); try lia; auto.
      ins; rewrite WMAX in *; rewrite H0; vauto.
    * specialize (WL_RMW n H (loc_l l)). rewrite E in WL_RMW; ins.
      apply WL_RMW.
      forward eapply I_HELPER with (k:=i) (j:=n); try lia; auto.
      unfold w_max; basic_solver.
  + destruct (nth n ol' init_ps (loc_l l)) eqn:E.
    f_equal.
    apply I1 in l0.
    destruct rmw_flag; desf.
    exfalso.
    red in SMATCH; desc.
    eapply wf_wl_last_max in WL_WF; try edone.
    2: by rewrite E; desf.
    eauto.
  + destruct (nth n ol' init_ps (loc_l l)) eqn:E.
    f_equal.
    apply I1 in l0.
    destruct rmw_flag; desf.
    exfalso.
    red in SMATCH; desc.
    eapply wf_wl_last_max in WL_WF; try edone.
    2: by rewrite E; desf.
    eauto.
- exists (fun n => upd (wl' (n + i)) x (wSeq (n + i))).

assert (HH: SGmatch_wl G t (skipn i ol)
  (fun n : nat => upd (wl' (n + i)) x (wSeq (n + i)))).


  { unfold ol. clear LEN' ol.
  rewrite skipn_app.
  autorewrite with calc_length.
  replace (min i (length ol')) with i by lia.
  rewrite skipn_mk_list; try lia.
  2: apply init_ps.
  rewrite Nat.sub_diag; ins.
  red; splits; red; splits; ins; autorewrite with calc_length.
  + unfold event_in_wl, in_wl, upd.
    unfolder; ins; desf.
    * destruct (classic (n + i = pred (length ol'))); ins; desf.
      -- rewrite H0.
         unfold w_max in LAST_WSEQ; unfolder in LAST_WSEQ; desf.
      -- eapply WSEQ with (n:= n+i). lia.
    * apply SMATCH.
      apply event_in_wl_k; lia.
  + unfold event_in_wl, in_wl, upd.
    unfolder; ins; desf.
    * destruct (classic (n + i = pred (length ol'))); ins; desf.
      -- rewrite H0.
         unfold w_max in LAST_WSEQ; unfolder in LAST_WSEQ; desf.
      -- eapply WSEQ with (n:= n+i). lia.
    * cut (acts_set G' (wl' (n + i) x)).
      { ins; eapply add_acts_set in H0; try edone.
        unfolder in H0; desf. exfalso. apply n0.
      red in SMATCH; desc. 
      red in WL_WF; desc.
      erewrite <- WL_LOC with (x:=x).
      rewrite <- H0. ins. lia. }
      apply SMATCH.
      apply event_in_wl_k; lia.
  + rewrite nth_mk_list; desc.
    remember (ifP n = pred (length ol') then true else false) as rmw_flag.
    clear Heqrmw_flag.
    autorewrite with calc_length in *; desf; try lia.
    unfold upd; desf.
    red in SMATCH; desc; red in WL_WF; desc.
    by rewrite <- WL_TID; [|lia].
    unfold upd; desf.
    red in SMATCH; desc; red in WL_WF; desc.
    by rewrite <- WL_TID; [|lia].
  + autorewrite with calc_length in *; desf; try lia.
    unfold upd; desf.
    * destruct (classic (n + i = pred (length ol'))); ins; desf.
      rewrite H.
      unfold w_max in LAST_WSEQ; unfolder in LAST_WSEQ; desf.
      cut (max_w_before_set G (⋃₁k <= (n + i), Wsets k) (loc_l l) (wSeq (n + i))).
      { unfold max_w_before_set. unfold max_elt. unfolder; ins; desf. }
      eapply WSEQ. lia.
    * red in SMATCH; desc; red in WL_WF; desc.
      by rewrite WL_LOC; [|lia].
  + rewrite nth_mk_list; desc.
    remember (ifP n = pred (length ol') then true else false) as rmw_flag.
    clear Heqrmw_flag.
    autorewrite with calc_length in *; desf; try lia.
    unfold upd; desf.
    red in SMATCH; desc; red in WL_WF; desc.
    by rewrite <- WL_VAL; [|lia].
    unfold upd; desf.
    red in SMATCH; desc; red in WL_WF; desc.
    by rewrite <- WL_VAL; [|lia].
  + autorewrite with calc_length in *; desf; try lia.
    rewrite nth_mk_list; desf; try lia.
    * unfold upd; desf.
      -- ins; desf; split; eauto; ins.
         apply w0.
      -- red in SMATCH; desc; red in WL_WF; desc.
         rewrite WL_RMW; [|lia].
         rewrite add_write_co; try edone.
         unfold add_co.
         split;
         unfolder; ins; desf; eauto 20.
         red in ADD; desf; ins; desf.
         eapply n0. rewrite <- LOCw.
         rewrite WL_LOC; auto; try lia.
         apply n0.
         erewrite <- WL_LOC with (x := x1); auto; try lia.
         rewrite <- REL; vauto.
         lia.
    * unfold upd; desf.
      -- ins.
         split; ins; eauto.
         exfalso.
         destruct (classic (n + i = pred (length ol'))).
         by rewrite H0 in n0; eauto. 
         forward apply WSEQ with (n:=n+i); try lia.
         unfold w_max in n0; unfolder in n0.
         unfold max_w_before_set; unfold max_elt; unfolder; ins; desf; eauto 10.
      -- red in SMATCH; desc; red in WL_WF; desc.
         rewrite WL_RMW; [|lia].
         rewrite add_write_co; try edone.
         unfold add_co.
         split;
         unfolder; ins; desf; eauto 20.
         red in ADD; desf; ins; desf.
         eapply n1. rewrite <- LOCw.
         rewrite WL_LOC; auto; try lia.
         apply n1.
         erewrite <- WL_LOC with (x := x1); auto; try lia.
         rewrite <- REL; vauto.
         lia.
  + unfold event_in_wl, in_wl, upd; desc; unfolder; ins; desc; subst.
    assert (HB: (hb G)^? z y) by basic_solver; clear H1.
    desf.
    * destruct (classic (n + i = pred (length ol'))).
      eapply LAST_WSEQ.
      rewrite H1 in H0.
      eapply H0.
      eapply (WSEQ (n+i)).
      lia.
      unfolder; ins; desf; splits; eauto.
      eexists _,y; splits; try (revert HB; basic_solver).
      exists 0; splits; try lia.
      unfold Wsets. 
      unfolder; desf.
      left.
      revert HB; basic_solver 12.
    * red in SMATCH; desc; red in WL_CONS; desc.
      eapply WL_O.
      unfolder; ins; splits.
      exists (wl' (n + i) x2).
      splits.
      apply event_order_in_wl_k with (k1:= i) (k2:= n+i); try lia.
      eexists; splits.
      eby eapply add_co_monotone.
      right.
      eapply add_hb with (G:=G); try edone.
      unfolder; ins; desf.
      apply WMAX in I0.
      rewrite I0.
      apply (wf_coE WF_G) in H0; revert H0.
      hahn_rewrite (wf_hbE WF_G) in HB; revert HB.
      basic_solver 21.
  + unfold event_order_in_wl, worder_wl, upd; desc; unfolder; ins; desc; subst.
    assert (HB: (hb G)^? z0
      (ifP x2 = (loc_l l) then wSeq (n1 + i) else wl' (n1 + i) x2)) by basic_solver; clear H1.
    desf; cycle 3.
    * red in SMATCH; desc; red in WL_CONS; desc. 
      rewrite <- add_co_monotone in WL_O; try edone.
      rewrite <- add_hb_monotone in WL_O; try edone.
      eapply WL_O.
      unfold event_order_in_wl, worder_wl.
      unfolder in HB.
      assert (n2 + i < length ol') by lia.
      assert (n1 + i <= n2 + i) by lia.
      basic_solver 21.
    * destruct (classic (n2 + i = pred (length ol'))).
      eapply LAST_WSEQ.
      rewrite H1 in H0.
      eapply H0.
      eapply WSEQ with (n := n2 + i) (b := z0); try lia.
      exists z0; splits; auto.
      unfold eqv_rel; splits; auto.
      assert (LT : n1 + i < pred (length ol')) by lia.
      apply WSEQ in LT; unfold max_w_before_set in LT.
      unfold "∩₁" in LT; desc.
      clear - LT1 HB H.
      revert HB LT1.
      assert (forall k, k <= n1 + i -> k <= n2 + i).
      lia.
      unfolder.
      ins; desf; generalize (@hb_trans G); try basic_solver 21.
    * destruct (classic (n1 + i = pred (length ol'))).
      { replace (n2 + i) with (pred (length ol')) in * by lia.
        red in SMATCH; desc; red in WL_WF; desc. 
        eapply WL_RMW with (n:= pred (length ol')).
        lia.
        eapply RMW_END; edone.
        eapply add_co_monotone; eauto. }
      assert (LT : n1 + i < pred (length ol')) by lia.
      specialize (WSEQ _ LT); try lia.
      unfold max_w_before_set in WSEQ.
      unfold "∩₁" in WSEQ; desc.
      assert (WSEQ1' : dom_rel ((hb G)^? ⨾ ⦗⋃₁k <= n1 + i, Wsets k⦘) z0).
      by revert HB WSEQ1; generalize (@hb_trans G); 
        clear; basic_solver 12.
      clear HB WSEQ1.
      unfold Wsets in WSEQ1'; desc.
      unfolder in WSEQ1'; desc; subst.
destruct WSEQ1'2; desc; subst; cycle 1.

{
rename y0 into w.
      red in SMATCH; desc; red in WL_CONS; desc. 
rewrite add_max_write_co_hb with (G:=G) (G':=G') in WL_O.
         5: try edone.
         all: try edone.
         eapply WL_O.
         eexists; splits.
         eapply event_order_in_wl_k with (k1:= i) (k2:=n2+i) (y:=y) (x:=(loc_l l)); try lia.
right.
unfolder; splits; eauto.
2: by symmetry; apply WMAX; auto.
right; splits; eauto. }


      assert (WSEQ1'' : (hb G)^? z0 y2) by basic_solver.
clear WSEQ1'.

      red in SMATCH; desc; red in WL_CONS; desc. 
      desf.
      -- rewrite add_max_write_dom_co_hb with (G:=G) (G':=G') in WL_D.
         5: try edone.
         all: try edone.
         eapply WL_D with (x:=(wl' (n2 + i) y)).
         split.
         by apply event_in_wl_k; lia.
         left; revert WSEQ1'' H0; basic_solver 21.
      -- eapply WL_O with (x := (wl' i (loc (wSeq (n1 + i))))).
         exists (wl' (n2 + i) y); splits.
         apply event_order_in_wl_k; try lia.
         exists z0; splits.
         by eapply add_co_monotone; eauto.
         right.
         eapply add_hb with (G:=G) (G':=G'); try edone.
         left. right.
         assert (acts_set G y2).
         { apply (wf_coE WF_G) in H0.
           hahn_rewrite (wf_hbE WF_G) in WSEQ1''.
           revert WSEQ1'' H0. basic_solver 21. }
         unfolder; ins; desf; splits; eauto 12.
         symmetry.
         apply WMAX. congruence.
      -- unfolder in WSEQ1''; desf.
         apply (wf_co_init WF_G) in H0.
         revert H0; basic_solver.
         apply (no_hb_to_init WF_G) in WSEQ1''.
         revert WSEQ1''; basic_solver.
      -- eapply WL_O.
         eexists; splits.
         eapply event_order_in_wl_k with (k1:= y1) (k2:=n2+i); try lia.
         eexists; splits.
         by eapply add_co_monotone; eauto.
         revert WSEQ1''; generalize (add_hb_monotone ADD); basic_solver.
    * destruct (classic (n2 + i = pred (length ol'))).
      eapply LAST_WSEQ.
      rewrite H1 in H0.
      eapply H0.
      eapply WSEQ with (n := (n2 + i)) (b:=z0).
      lia. splits; eauto.
      assert (n1+i <= n2+i) by lia.
      assert (Wsets (n1 + i) (wl' (n1 + i) x2)).
      { unfold Wsets. left. basic_solver 21. }
      revert HB; basic_solver 21. }

  splits; try done.


  ins. rewrite upds.
unfold max_elt; ins.

destruct (classic (i = Init.Nat.pred (length ol'))).
{ subst; eapply I0.
exfalso. eby eapply LAST_WSEQ. }


eapply WSEQ with (n:=i) (b:=w).
lia.
eexists w.
splits.
+ 
assert (A: i < Init.Nat.pred (length ol')) by lia.
specialize (WSEQ i A). clear A.
red in WSEQ; unfold "∩₁" in WSEQ; desc.
forward (eapply max_w_maximum with (G:=G) (wm:=w) (w:=(wSeq i))); try edone.
unfolder; ins; desf.
exfalso.
eby eapply MAX.
+ 
unfolder; splits; eauto.
eexists; splits; eauto.
exists 0; splits; [lia|].
 unfold Wsets.
basic_solver. 


- exists (fun n => ifP (n < i) then (wl' n) 
                   else upd (wl' n) x (wSeq n)).
unfold ol.
red; splits; red; splits; ins; autorewrite with calc_length.
  + unfold event_in_wl, in_wl, upd.
    unfolder; ins; desf.
    * apply SMATCH.
      apply event_in_wl_k. lia.
    * destruct (classic (n = pred (length ol'))); ins; desf.
      apply LAST_WSEQ. 
      apply WSEQ. lia.
    * apply SMATCH.
      apply event_in_wl_k. lia.
  + unfold event_in_wl, in_wl, upd.
    unfolder; ins; desf.
    * cut (acts_set G' (wl' n x)).
      { ins; eapply add_acts_set in H0; try edone.
        unfolder in H0; desf. exfalso. 
        eapply I1; try edone.
        apply WMAX.
        rewrite H0.
        f_equal.
        red in SMATCH; desc. red in WL_WF; desc.
        erewrite <- WL_LOC with (x:=x).
        rewrite <- H0.
        vauto.
        lia. }
      apply SMATCH.
      apply event_in_wl_k. lia.
    * destruct (classic (n = pred (length ol'))); ins; desf.
      apply LAST_WSEQ. 
      apply WSEQ. lia.
    * cut (acts_set G' (wl' n x)).
      { ins; eapply add_acts_set in H0; try edone.
        unfolder in H0; desf. exfalso. apply n1.
        red in SMATCH; desc. 
        red in WL_WF; desc.
        erewrite <- WL_LOC with (x:=x).
        rewrite <- H0. ins.
        lia. }
      apply SMATCH.
      apply event_in_wl_k. lia.
  + rewrite nth_app; desf;
    autorewrite with calc_length in *; 
    rewrite !nth_mk_list; unfold upd; desf; ins; try lia.
    all: try replace (n - i + i) with n in * by lia; auto.
    all: try by (apply SMATCH; lia).
  + autorewrite with calc_length in *; 
    unfold upd; desf.
    all: try by (apply SMATCH; lia).
    destruct (classic (n = pred (length ol'))); ins; desf.
    by apply LAST_WSEQ.
    apply WSEQ; lia.
  + rewrite nth_app; desf;
    autorewrite with calc_length in *; 
    rewrite !nth_mk_list; unfold upd; desf; ins; try lia.
    all: try replace (n - i + i) with n in * by lia; auto.
    all: try by (apply SMATCH; lia).
  + rewrite nth_app; ins;
    autorewrite with calc_length in *; 
    rewrite !nth_mk_list; unfold upd; desc; ins; try lia.
    destruct (classic (x1=x)); subst.
    * desf; try lia.
      all: try replace (n - i + i) with n in * by lia; auto.
      all: ins; try lia; split; ins; eauto.
      all: try apply w0.
      -- destruct (classic (n = pred (length ol'))); ins; desf.
         exfalso.
         apply n0; unfold w_max; unfolder; splits; eauto.
         apply WSEQ. lia.
         apply WSEQ. lia.
         apply WSEQ. lia.
      -- rewrite add_co_monotone; try edone.
         apply SMATCH; try lia; auto.
      -- exfalso.
         apply n0.
         unfold w_max; unfolder; splits; eauto.
         cut (acts_set G' (wl' n (loc_l l))).
         { ins; eapply add_acts_set in H0; try edone.
           unfolder in H0; desf. exfalso.
           eapply I1 with (j:=n).
           2: apply WMAX; symmetry.
           lia. auto. }
         apply SMATCH; apply event_in_wl_k; lia.
         apply SMATCH; apply event_in_wl_k; lia.
         apply SMATCH; lia.
    * red in SMATCH; desc; red in WL_WF; desf.
      all: try replace (n - i + i) with n in * by lia; auto.
      all: try rewrite WL_RMW; auto; try lia.
      all: split.
      all: try eby rewrite <- add_co_monotone.
      all: rewrite add_max_write_co with (G:=G) (G':=G') (w:=w); try edone; vauto.
      all: assert (LL: ~ is_loc (loc_l l) (wl' n x1))
      by by unfold is_loc; rewrite WL_LOC; try lia; auto.
      all: unfolder; ins; desf; eauto.
  + unfold event_in_wl, in_wl, upd; desc; unfolder; ins; desc; subst.
    assert (HB: (hb G)^? z y) by basic_solver; clear H1.
    destruct (classic (x2=(loc_l l))); subst.
    * desf.
      -- eapply add_co_monotone in H0; try edone.
         assert (HB' : (hb G')^? z y).
         { clear H0. unfolder in HB; desf. basic_solver.
           right. eapply add_hb_monotone; try edone. }
         clear HB.
         red in SMATCH; desc; red in WL_CONS; desc.
         eapply WL_D; unfolder; splits; eauto 12.
         desf; apply event_in_wl_k; lia.
      -- destruct (classic (n = pred (length ol'))).
         by eapply LAST_WSEQ; rewrite H1 in H0; eapply H0.
         eapply (WSEQ n).
         lia.
         unfolder; ins; desf; splits; eauto.
         eexists _,y; splits; try (revert HB; basic_solver).
         exists 0; splits; try lia.
         unfold Wsets. 
         unfolder; desf.
         left.
         revert HB. basic_solver 12.
    * eapply add_co_monotone in H0; try edone.
      assert (HB' : (hb G')^? z y).
      { clear H0. unfolder in HB; desf. basic_solver.
      right.
      eapply add_hb_monotone; try edone. }
      clear HB.
      red in SMATCH; desc; red in WL_CONS; desc.
      eapply WL_D; unfolder; splits; eauto 12.
      desf; apply event_in_wl_k; lia.
  + replace (i + (length ol' - i)) with (length ol') by lia.
    unfold event_order_in_wl, worder_wl, upd; desc; unfolder; ins; desc; subst.
    destruct (classic (x2=(loc_l l) /\ n1 >= i)); desc; subst.
    * assert (HB : (hb G)^? z0 (wSeq n1)). 
      by clear H0; unfolder; desf; try lia; auto.
      clear H1. 
      destruct (classic (y=(loc_l l))); desc; subst.
      -- desf; try lia.
         destruct (classic (n2 = pred (length ol'))); subst.
         eby eapply LAST_WSEQ.
         eapply WSEQ with (n := n2) (b := z0); try lia.
         exists z0; splits; auto.
         unfold eqv_rel; splits; auto.
         assert (LT : n1 < pred (length ol')) by lia.
         apply WSEQ in LT; unfold max_w_before_set in LT.
         unfold "∩₁" in LT; desc.
         clear - LT1 HB H.
         revert HB LT1.
         assert (forall k, k <= n1 -> k <= n2).
         lia.
         unfolder.
         ins; desf; generalize (@hb_trans G); try basic_solver 21.
      -- desf; try lia.
         destruct (classic (n1 = pred (length ol'))).
         { replace (n2) with (pred (length ol')) in * by lia.
           red in SMATCH; desc; red in WL_WF; desc. 
           eapply WL_RMW with (n:= pred (length ol')).
           lia.
           eapply RMW_END; edone.
           eapply add_co_monotone; eauto. }
         assert (LT : n1  < pred (length ol')) by lia.
         specialize (WSEQ _ LT); try lia.
         unfold max_w_before_set in WSEQ.
         unfold "∩₁" in WSEQ; desc.
         assert (WSEQ1' : dom_rel ((hb G)^? ⨾ ⦗⋃₁k <= n1, Wsets k⦘) z0).
         by revert HB WSEQ1; generalize (@hb_trans G); clear; basic_solver 12.
         clear HB WSEQ1.
         unfold Wsets in WSEQ1'; desc.
         unfolder in WSEQ1'; desc; subst.


destruct WSEQ1'2; desc; subst; cycle 1.

{
rename y0 into w.
      red in SMATCH; desc; red in WL_CONS; desc. 
rewrite add_max_write_co_hb with (G:=G) (G':=G') in WL_O.
         5: try edone.
         all: try edone.
         eapply WL_O.
         eexists; splits.
         eapply event_order_in_wl_k with (k1:= i) (k2:=n2) (y:=y) (x:=(loc_l l)); try lia.
right.
unfolder; splits; eauto.
2: by symmetry; apply WMAX; auto.
right; splits; eauto. }


         assert (WSEQ1'' : (hb G)^? z0 y2) by basic_solver.
         clear WSEQ1'.
         red in SMATCH; desc; red in WL_CONS; desc. 
         desf.
         ++ rewrite add_max_write_dom_co_hb with (G:=G) (G':=G') in WL_D.
            5: try edone.
            all: try edone.
            eapply WL_D with (x:=(wl' (n2) y)).
            split.
            by apply event_in_wl_k; lia.
            left; revert WSEQ1'' H0; basic_solver 21.
         ++ eapply WL_O with (x := (wl' i (loc (wSeq (n1))))).
            exists (wl' (n2) y); splits.
            apply event_order_in_wl_k; try lia.
            exists z0; splits.
            by eapply add_co_monotone; eauto.
            right.
            eapply add_hb with (G:=G) (G':=G'); try edone.
            left; right.
            assert (acts_set G y2).
            { apply (wf_coE WF_G) in H0.
            hahn_rewrite (wf_hbE WF_G) in WSEQ1''.
            revert WSEQ1'' H0. basic_solver 21. }
            unfolder; ins; desf; splits; eauto 12.
            symmetry.
            apply WMAX. congruence.
         ++ unfolder in WSEQ1''; desf.
            apply (wf_co_init WF_G) in H0.
            revert H0; basic_solver.
            apply (no_hb_to_init WF_G) in WSEQ1''.
            revert WSEQ1''; basic_solver.
         ++ eapply WL_O.
            eexists; splits.
            eapply event_order_in_wl_k with (k1:= y1) (k2:=n2); try lia.
            eexists; splits.
            by eapply add_co_monotone; eauto.
            revert WSEQ1''; generalize (add_hb_monotone ADD); basic_solver.
      * eapply add_co_monotone in H0; try edone.
        assert (HB : (hb G)^? z0 (wl' n1 x2)). 
        { clear H0. 
          unfolder; desf; try basic_solver.
          exfalso; eapply H2; splits; auto; lia.
          exfalso; eapply H2; splits; auto; lia. }
        clear H1.
        destruct (classic (y=(loc_l l) /\ n2 >= i)); desc; subst.
        -- eapply add_max_write_co with (G:=G) (G':=G') in H0.
           5: edone. all: try edone.
           desf; try lia.
           unfolder in H0; desf.
           ++ destruct (classic (n2 = pred (length ol'))); subst.
              eby eapply LAST_WSEQ.
              eapply WSEQ with (n:=n2) (b:=z0).
              lia.
              exists z0; splits; eauto.
              unfolder. splits; ins; desf; eauto.
              exists (wl' n1 x2); splits.
              revert HB; basic_solver.
              exists n1; splits; [lia|].
              unfold Wsets; ins. left; basic_solver 12.
          ++ hahn_rewrite (wf_hbE WF_G) in HB; unfolder in HB; desf.
             ** replace x2 with (loc_l l) in *.
                eapply I1 with (j:=n1).
                assert (~ n1 >= i) by tauto; lia.
                by apply WMAX; auto.
                red in SMATCH; desc; red in WL_WF; desc.
                erewrite <- WL_LOC with (x:=x2).
                erewrite <- HB; vauto.
                lia.
             ** eapply add_fresh; try edone.
        -- assert (CO : (co G') (wl' n2 y) z0). 
           { generalize (add_co_monotone ADD).
             unfolder; desf; ins; try basic_solver.
             exfalso; eapply H1; splits; eauto. lia. }
           clear H0 H1.
           red in SMATCH; desc; red in WL_CONS; desc.
           eapply WL_O; unfolder. 
           eexists (wl' n2 y); splits.
           eapply event_order_in_wl_k; [lia|edone].
           generalize (add_hb_monotone ADD).
           revert HB; basic_solver 12.

Unshelve.
apply init_ps.
apply default_event.
Qed.


Lemma LsimG_write (st': State) (G G': execution) 
          t x v (SRAG: SRAG_step G G' (Some (t, (Astore x v))))
          (MATCH': SGmatch st' G') : 
          exists st, SRAL_step st st' (Some (t, (Astore x v))) /\ 
                     SGmatch st G.
Proof.
forward eapply LsimG_write_main; try edone.
ins; desc.
eexists; splits; [econs; eauto| edone].
Qed.
